import os
import logging
import shutil

from typing import (
    List,
    Optional,
    Union,
)

from dataclasses import dataclass

from pathlib import Path

import tensorflow as tf
import transformers

from transformers import (
    PretrainedConfig,
    PreTrainedTokenizerFast,
    TFPreTrainedModel,
    TFAutoModelForMaskedLM,
    AutoConfig,
    AutoTokenizer,
)

from tokenizers.pre_tokenizers import Whitespace

from nlpipes.data.data_loaders import DataLoader
from nlpipes.data.data_selectors import DataSelector
from nlpipes.data.data_augmentors import VocabAugmentor
from nlpipes.data.data_processors import DataProcessorForSequenceClassification
from nlpipes.data.data_cleaners import clean

from nlpipes.trainers.trainers import Trainer

from nlpipes.losses.losses import softmax_cross_entropy

from nlpipes.data.data_utils import (
     create_examples,
     split_examples,
     generate_batches,
     chunk,
)

from nlpipes.pipelines.pipeline_utils import (
     preprocess,
     tokenize,
     encode,
     estimate,
     postprocess,
)

from nlpipes.data.data_types import (
     Corpus,
     Document,
     Token,
     InputExample,
     InputTokens,
     InputFeatures,
     Predictions,
     Outcomes,
)

from nlpipes.callbacks.callbacks import (
     Callback,
     TrainingStep,
     History,
     ModelCheckpoint,
     CSVLogger,
     ProgbarLogger,
     EarlyStopping,
     TimedStopping,
)

transformers.logging.set_verbosity_error()

logger = logging.getLogger('__name__')

#uncomment row below for debugging rust tokenizer
#os.environ["RUST_BACKTRACE"]="full" 


@dataclass
class TFPipelineForSequenceClassification():
    """ The pipeline for mono-label sequence classification """
    
    model: TFPreTrainedModel
    tokenizer: PreTrainedTokenizerFast
    config: PretrainedConfig
    
    def train(self, 
              X: List[str], 
              Y: List[str],
              train_frac: float = 0.75,
              test_frac: float = 0.25,
              num_epochs: int = 3,
              batch_size: int = 4,
              lr: float = 2e-5,
              patience: int = 5,
              min_delta: float = 0.01,
              time_limit: int = 86400,
              rand_seed: int = 69,
              keep_checkpoints: bool = False,
              checkpoints_dir: str = './checkpoints',
              logs_dir: str = './logs',
             ) -> List[History]:
        """ Train a model for mono-label classification.
        
        Most of the behavior of the training process is implemented
        in callbacks objects. Even the actual training step 
        (calculating the gradient and updating the weights using 
        an optimizer) is implemented as callbacks and is not part of
        the Trainer itself. 
        This modular architecture allows new ideas to be implemented 
        without having to change and increase the complexity of the
        trainer. This keeps the overall system clean, even if many 
        different training improvements are implemented """
        
        examples = create_examples(
            texts=X, 
            labels=Y,
            config=self.config,
        )
        
        train_examples, test_examples = split_examples(
            examples,
            train_frac=train_frac,
            test_frac=test_frac,
            shufffle=True,
        )
        
        data_processor = DataProcessorForSequenceClassification(
            name='DataProcessing',
            tokenizer=self.tokenizer,
            config=self.config,
        )
        
        train_stream = DataLoader(
            name='TrainDataLoader',
            examples=train_examples,
            batch_size=batch_size,
            data_processor=data_processor,
        )
        
        test_stream = DataLoader(
            name='TestDataLoader',
            examples=test_examples,
            batch_size=batch_size,
            data_processor=data_processor,
        )
        
        training_step = TrainingStep(
            name='TrainingStep',
            model=self.model,
            loss_function=softmax_cross_entropy,
            optimizer=tf.optimizers.Adam(lr),
        )
        
        history = History(
            name='History', 
            training_step=training_step,
            metric=tf.metrics.Mean,
        )
        
        model_checkpoint = ModelCheckpoint(
            name='ModelCheckpoint',
            model=self.model, 
            history=history,
            checkpoints_dir=checkpoints_dir,
        )
        
        early_stopping = EarlyStopping(
            name='EarlyStopping', 
            history=history,
            direction='minimize',
            min_delta=min_delta,
            patience=patience,
        )
        
        timed_stopping = TimedStopping(
            name='TimedStopping',
            time_limit=time_limit,
        )
       
        csv_logger = CSVLogger(
            name='CSVLogging',
            logs_dir=logs_dir,
        )
        
        progbar_logger = ProgbarLogger(
            name='ProgbarLogger',
            batch_size=batch_size,
            num_epochs=num_epochs,
            history=history,
            num_samples=sum([
                train_stream.num_examples(),
                test_stream.num_examples(),
            ])
        )
        
        callbacks = [
            training_step,
            history,
            model_checkpoint,
            csv_logger,
            progbar_logger,
            early_stopping,
            timed_stopping,
        ]
        
        trainer = Trainer(
            train_stream=train_stream,
            test_stream=test_stream,
            callbacks=callbacks,
            num_epochs=num_epochs,
        )
        
        trainer.train()
        
        if keep_checkpoints == False:
            shutil.rmtree(checkpoints_dir)
         
        return history
    
    def evaluate(self,
                 X: List[str],
                 Y: List[str],
                 metric: tf.metrics.Metric,
                 batch_size: int = 32,
                ) -> tf.metrics.Metric:
        """ Evaluate the model performance on a new unseen 
        labeled data according to the user-defined evaluation
        metric """
        
        for texts, labels in generate_batches(X, Y, batch_size):
            examples = preprocess(texts, labels)
            tokens = tokenize(examples, self.tokenizer)
            features = encode(tokens, self.tokenizer)
            predictions = estimate(features, self.model, self.config)
            
            y_pred = predictions.label
            y_true = [example.label for example in examples]
            y_true = [self.config.label2id[y] for y in y_true]
            
            metric.update_state(y_true, y_pred)
        
        return metric.result()
    
    def predict(self, 
                texts: Corpus, 
                chunk_text: bool = False,
               ) -> Outcomes:
        """ Get predictions a new unseen unlabeled data """
        
        examples = preprocess(texts)
        tokens = tokenize(examples, self.tokenizer)
        features = encode(tokens, self.tokenizer)
        predictions = estimate(features, self.model, self.config)
        outcomes = postprocess(predictions, self.config)
        
        return outcomes
    
    def save(self, save_dir: str):
        """ Save the trained model on disk """
        
        os.makedirs(save_dir, exist_ok=True)
        self.model.save_pretrained(save_dir)
        self.tokenizer.save_pretrained(save_dir)
        self.config.save_pretrained(save_dir)
    


from transformers import PretrainedConfig

        
class AdaptedBertConfig(PretrainedConfig):
    """ This is the configuration class to store the configuration 
    of a `AdaptedBertModel`. It is used to instantiate a BERT model
    with additionnal adpaters modules according to the specified 
    arguments, defining the model architecture. """
    
    model_type = "bert"

    def __init__(
        self,
        vocab_size=30522,
        hidden_size=768,
        num_hidden_layers=12,
        num_attention_heads=12,
        intermediate_size=3072,
        hidden_act="gelu",
        hidden_dropout_prob=0.1,
        attention_probs_dropout_prob=0.1,
        max_position_embeddings=512,
        type_vocab_size=2,
        initializer_range=0.02,
        layer_norm_eps=1e-12,
        pad_token_id=0,
        position_embedding_type="absolute",
        use_cache=True,
        classifier_dropout=None,
        bottleneck_size=64,
        non_linearity='gelu_new',
        **kwargs
    ):
        super().__init__(pad_token_id=pad_token_id, **kwargs)

        self.vocab_size = vocab_size
        self.hidden_size = hidden_size
        self.num_hidden_layers = num_hidden_layers
        self.num_attention_heads = num_attention_heads
        self.hidden_act = hidden_act
        self.intermediate_size = intermediate_size
        self.hidden_dropout_prob = hidden_dropout_prob
        self.attention_probs_dropout_prob = attention_probs_dropout_prob
        self.max_position_embeddings = max_position_embeddings
        self.type_vocab_size = type_vocab_size
        self.initializer_range = initializer_range
        self.layer_norm_eps = layer_norm_eps
        self.position_embedding_type = position_embedding_type
        self.use_cache = use_cache
        self.classifier_dropout = classifier_dropout
        self.bottleneck_size = bottleneck_size
        self.non_linearity = non_linearity



from typing import Callable
from typing import List

from dataclasses import dataclass

import copy

import numpy as np
import tensorflow as tf

from transformers import (
    PretrainedConfig,
    PreTrainedTokenizer,
    AutoTokenizer,
)

from nlpipes.data.data_types import InputExample
from nlpipes.data.data_types import InputFeatures


""" The data processor classes.

The data processor encapsulates all the transformations
needed to encode input examples into input features 
expected by the model. """


@dataclass
class DataProcessorForLanguageModeling:
    
    """ 
    Data processor for Masked Language Modeling:
    80% MASK, 10% random, 10% original.
    Inputs are dynamically padded to the maximum 
    length of a batch if they are not all of the 
    same length.
    
    Args
    ----------

    tokenizer (transformers.PreTrainedTokenizer):
        The tokenizer used for encoding the data.

    config (transformers.PreTrainedCOnfig):
        The configuration of the model.

    """
    name: str = 'LanguageModelingProcessor'
    tokenizer: PreTrainedTokenizer = None
    config: PretrainedConfig = None
    mlm_proba: float = None
    
    def __call__(self, 
                 examples: List[InputExample],
                ) -> InputFeatures:
        
        texts = [example.text for example in examples]
        
        inputs = self.tokenizer(
            texts,
            add_special_tokens=True,
            padding=True,
            truncation=True,
            max_length=512,
            is_split_into_words=False,
            return_tensors='tf',
        )
        
        input_ids = tf.identity(inputs.input_ids)
        
        mask = self.create_mask(
            inputs=inputs.input_ids,
            mask_frac=self.mlm_proba,
            excluded=self.tokenizer.all_special_ids,
        )
        
        inputs['input_ids'] = self.apply_mask(
            inputs=inputs.input_ids,
            mask=mask,
            mask_token_id=self.tokenizer.mask_token_id,
        )
        
        labels = self.apply_mask(
            inputs=input_ids,
            mask=~mask,
            mask_token_id=-100,
        )
        
        inputs['label'] = tf.one_hot(
            [label for label in labels],
            depth=len(self.tokenizer)
        )
        
        input_features = InputFeatures(
            input_ids=inputs['input_ids'],
            attention_mask=inputs['attention_mask'],
            token_type_ids=inputs['token_type_ids'],
            label=inputs['label'],
        )
        
        return input_features
    
    def create_mask(
        self,
        inputs: tf.Tensor,
        mask_frac: float,
        excluded: tf.Tensor,  
    ) -> tf.Tensor:
        """ Create the mask used for the masked 
        language modeling """
        
        random = tf.random.uniform(inputs.shape)
        
        mask = tf.math.logical_and(
            random < mask_frac, 
            ~np.isin(inputs, excluded),
        )
        
        return mask
    
    def apply_mask(
        self,
        inputs: tf.Tensor,
        mask: tf.Tensor,
        mask_token_id: int,
    ) -> tf.Tensor:
        """ Apply the mask to the given inputs """
        
        masked_inputs = tf.where(
            condition=mask,
            x=mask_token_id,
            y=inputs,
        )
        
        return masked_inputs


@dataclass
class DataProcessorForSequenceClassification:
    
    """
    Data Processor for Sequence Classification.
    Inputs are dynamically padded to the maximum 
    length of a batch if they are not all of the 
    same length.
    
    Args
    ----------

    tokenizer (transformers.PreTrainedTokenizer):
        The tokenizer used for encoding the data.

    config (transformers.PreTrainedCOnfig):
        The configuration of the model.

    """
    name: str = 'SequenceClassificationProcessor'
    tokenizer: PreTrainedTokenizer = None
    config: PretrainedConfig = None
    
    def __call__(self, 
                 examples: List[InputExample],
                ) -> InputFeatures:
        
        texts = [example.text for example in examples]
        
        inputs = self.tokenizer(
            texts,
            add_special_tokens=True,
            padding=True,
            truncation = True,
            max_length=512,
            is_split_into_words = False,
            return_tensors='tf',
        )
        
        inputs['label'] = tf.one_hot(
            [example.label for example in examples],
            depth=self.config.num_labels
        )
        
        input_features = InputFeatures(
            input_ids=inputs['input_ids'],
            attention_mask=inputs['attention_mask'],
            token_type_ids=inputs['token_type_ids'],
            label=inputs['label'],
        )
        
        return input_features
    


"""Similiarity metrics for in-domain subcorpus selection
Code from: 
reference: https://github.com/georgian-io/Transformers-Domain-Adaptation.
reference: https://github.com/sebastianruder/learn-to-select-data.
"""

from typing import Callable, Dict
from typing_extensions import Literal

import numpy as np
import scipy.spatial.distance
import scipy.stats


def jensen_shannon_similarity(repr1: np.ndarray, repr2: np.ndarray) -> np.ndarray:
    """Calculate similairty based on Jensen-Shannon divergence.
    ref: https://en.wikipedia.org/wiki/Jensen%E2%80%93Shannon_divergence
    """
    if len(repr1) == 1:
        repr1 = np.repeat(repr1, len(repr2), axis=0)
    elif len(repr2) == 1:
        repr2 = np.repeat(repr2, len(repr1), axis=0)
    avg_repr = 0.5 * (repr1 + repr2)
    sim = np.array(
        [
            1 - 0.5 * (scipy.stats.entropy(p, avg) + scipy.stats.entropy(q, avg))
            for p, q, avg in zip(repr1, repr2, avg_repr)
        ]
    )

    # the similarity is -inf if no term is in the vocabulary
    sim = np.where(np.isinf(sim), 0, sim)
    return sim

def renyi_similarity(
    repr1: np.ndarray, repr2: np.ndarray, alpha: float = 0.99
) -> np.ndarray:
    """Calculate similarity based on Rényi divergence.
    ref: https://en.wikipedia.org/wiki/R%C3%A9nyi_entropy#R.C3.A9nyi_divergence
    """
    log_sum = (np.power(repr1, alpha) / np.power(repr2, alpha - 1)).sum(axis=-1)
    renyi_divergence = 1 / (alpha - 1) * np.log(log_sum)
    return -renyi_divergence

def cosine_similarity(repr1: np.ndarray, repr2: np.ndarray) -> np.ndarray:
    """Calculate cosine similarity 
    ref: https://en.wikipedia.org/wiki/Cosine_similarity)."""
    if len(repr1) == 1:
        repr1 = np.repeat(repr1, len(repr2), axis=0)
    elif len(repr2) == 1:
        repr2 = np.repeat(repr2, len(repr1), axis=0)

    assert not (np.isnan(repr2).any() or np.isinf(repr2).any())
    assert not (np.isnan(repr1).any() or np.isinf(repr1).any())
    sim = np.array(
        [1 - scipy.spatial.distance.cosine(p, q) for p, q in zip(repr1, repr2)]
    )

    # the similarity is nan if no term in the document is in the vocabulary
    sim = np.where(np.isnan(sim), 0, sim)
    return sim

def euclidean_similarity(repr1: np.ndarray, repr2: np.ndarray) -> np.ndarray:
    """Calculate similarity based on Euclidean distance.
    ref: https://en.wikipedia.org/wiki/Euclidean_distance
    """
    euclidean_distance = np.sqrt(((repr1 - repr2) ** 2).sum(axis=-1))
    return -euclidean_distance

def variational_similarity(repr1: np.ndarray, repr2: np.ndarray) -> np.ndarray:
    """Calculate similarity based on L1 / Manhattan distance.
    ref: https://en.wikipedia.org/wiki/Taxicab_geometry
    """
    manhattan_distance = np.abs(repr1 - repr2).sum(axis=-1)
    return -manhattan_distance

def bhattacharyya_similarity(repr1: np.ndarray, repr2: np.ndarray) -> np.ndarray:
    """Calculate similarity based on Bhattacharyya distance.
    ref: https://en.wikipedia.org/wiki/Bhattacharyya_distance
    """
    distance = -np.log(np.sqrt(repr1 * repr2).sum(axis=-1))
    assert not np.isnan(distance).any(), "Error: Similarity is nan."

    # the distance is -inf if no term is in the vocabulary
    distance = np.where(np.isinf(distance), 0, distance)
    return -distance


SimilarityMetric = Literal[
    "jensen-shannon",
    "renyi",
    "cosine",
    "euclidean",
    "variational",
    "bhattacharyya",
]
SimilarityFunc = Callable[
    [np.ndarray, np.ndarray], np.ndarray
]
SIMILARITY_FEATURES = {
    "jensen-shannon",
    "renyi",
    "cosine",
    "euclidean",
    "variational",
    "bhattacharyya",
}


def similarity_metrics_factory(metric: SimilarityMetric) -> SimilarityFunc:
    """Return the corresponding similarity function based on the provided metric.
    Args:
        metric (str): Similarity metric
    Raises:
        ValueError: If `metric` does not exist in SIMILARITY_FEATURES
    """
    if metric not in SIMILARITY_FEATURES:
        raise ValueError(f'"{metric}" is not a valid similarity metric.')

    mapping: Dict[SimilarityMetric, SimilarityFunc] = {
        "jensen-shannon": jensen_shannon_similarity,
        "renyi": renyi_similarity,
        "cosine": cosine_similarity,
        "euclidean": euclidean_similarity,
        "variational": variational_similarity,
        "bhattacharyya": bhattacharyya_similarity,
    }

    similarity_function = mapping[metric]
    return similarity_function



# Code lightly adapted from https://github.com/georgian-io/Transformers-Domain-Adaptation

from collections import Counter

from pathlib import Path

from tempfile import NamedTemporaryFile
from tempfile import TemporaryDirectory

from types import MappingProxyType

from typing import IO
from typing import Counter as CounterType
from typing import List, Type, Union

from transformers import PreTrainedTokenizer

from tokenizers import Tokenizer, trainers
from tokenizers.implementations import BaseTokenizer
from tokenizers.models import BPE, Unigram, WordPiece
from tokenizers.normalizers import Lowercase
from tokenizers.pre_tokenizers import Whitespace

from nlpipes.data.data_types import Corpus, Token


class VocabAugmentor():
    """Find new tokens to add to the :obj:`tokenizer`'s vocabulary.
    A new vocabulary is learnt from the training corpus using the
    same tokenization model. The most common tokens of this new 
    vocabulary that do not exist in the existing vocabulary are 
    selected
    """

    supported_trainers = MappingProxyType(
        {
            BPE: trainers.BpeTrainer,
            WordPiece: trainers.WordPieceTrainer,
            Unigram: trainers.UnigramTrainer,
        }
    )

    def __init__(self,
                 name: str,
                 target_vocab_size: int,
                 tokenizer: PreTrainedTokenizer, 
                 cased: bool, 
    ):
        """
        Args:
            tokenizer: A pre-trained 🤗 Tokenizer
            cased: If False, ignore uppercases in corpus
            target_vocab_size: Size of augmented vocabulary
        Raises:
            ValueError: If :obj:`target_vocab_size` is larger or equal to the
                        existing vocabulary of :obj:`tokenizer`
            RuntimeError: If :obj:`tokenizer` uses an unsupported tokenization model
        """
        if target_vocab_size <= tokenizer.vocab_size:
            raise ValueError(
                f"Ensure that `target_vocab_size` is larger than tokenizer's vocab size."
            )
        self.target_vocab_size = target_vocab_size
        self.tokenizer = tokenizer
        self.cased = cased
        self.model_cls: Type[
            BaseTokenizer
        ] = tokenizer.backend_tokenizer.model.__class__

        # Instantiate rust tokenizer
        rust_tokenizer = Tokenizer(self.model_cls())
        if not cased:
            rust_tokenizer.normalizer = Lowercase()
        rust_tokenizer.pre_tokenizer = Whitespace()
        self.rust_tokenizer = rust_tokenizer

        # Instantiate the appropriate Trainer based on `self.model` (i.e. BPE, WordPiece, etc)
        trainer_cls = self.supported_trainers.get(self.model_cls, None)
        if trainer_cls is None:
            raise RuntimeError(f"{self.model_cls} is not supported")
        self.trainer = trainer_cls(
            vocab_size=self.target_vocab_size,
            special_tokens=list(self.tokenizer.special_tokens_map.values()),
        )

    def get_new_tokens(
        self,
        training_corpus: Union[Corpus, Path, str],
    ) -> List[Token]:
        """Obtain new tokens found in :obj:`training_corpus`.
        New tokens contains the most common tokens that do not exist in 
        the :obj:`tokenizer`'s vocabulary.
        Args:
            training_corpus: The training corpus
        """
        # Training has to be wrapped with the `tmpfile` context
        with NamedTemporaryFile("w+") as tmpfile:  # If we need to save Corpus type
            # Train new tokenizer on `ft_corpus`
            train_files = self._get_training_files(training_corpus, _tmpfile=tmpfile)
            
            self.rust_tokenizer.train(train_files, self.trainer)

            # Include unknown token to vocab
            with TemporaryDirectory() as tmpdir:
                files = self.rust_tokenizer.model.save(tmpdir)
                self.rust_tokenizer.model = self.model_cls.from_file(
                    *files, unk_token="[UNK]"
                )

            # Find most common tokens in vocab
            token_counts = self._count_tokens(train_files)
            
        # Remove overlapping tokens from original tokenizer
        token_counts = self._remove_overlapping_tokens(token_counts)
        
        new_tokens = [
            token
            for token, _ in token_counts.most_common(
                self.target_vocab_size - self.tokenizer.vocab_size
            )
        ]
        return new_tokens

    @staticmethod
    def _get_training_files(
        corpus: Union[Corpus, Path, str], _tmpfile: IO[str]
    ) -> List[str]:
        """Return files for training.
        If `corpus is a sequence of documents, it will be written to a temporary file,
        and that temporary file's name will be returned.
        If `corpus` is a Path or str, it will return the path, or paths if `corpus` is a directory.
        Args:
            corpus: Text data or path to training corpus
            _tmpfile: Temporary file object. Used when `corpus` is not a path
        Raises:
            FileNotFoundError: If `corpus` is a str or Path and it does not exist on the filesystem.
        """
        if isinstance(corpus, str) or isinstance(corpus, Path):
            corpus = Path(corpus)

            if not corpus.exists():
                raise FileNotFoundError(
                    f"Training corpus {corpus.as_posix()} does not exist."
                )

            files = list(corpus.rglob("*.*")) if corpus.is_dir() else [corpus]
            files = [f.as_posix() for f in files]
            return files

        else:  # Corpus type
            for doc in corpus:
                _tmpfile.write(doc)
            _tmpfile.seek(0)
            return [_tmpfile.name]

    def _count_tokens(self, files: List[str]) -> CounterType[str]:
        """Count number of tokens in a list of files."""
        token_counts: CounterType[str] = Counter()
        for file in files:
            with open(file) as f:
                token_counts += Counter(
                    token
                    for enc in self.rust_tokenizer.encode_batch(f.readlines())
                    for token in enc.tokens
                )
        return token_counts

    def _remove_overlapping_tokens(
        self, token_counts: CounterType[str]
    ) -> CounterType[str]:
        """Remove tokens from `token_counts` that exist in the current tokenizer's vocab."""
        _token_counts = token_counts.copy()
        
        for vocab_term in self.tokenizer.get_vocab().keys():
            if vocab_term in _token_counts:
                del _token_counts[vocab_term]
        return _token_counts


import numpy as np
import tensorflow as tf

from typeguard import typechecked


class ConfusionMatrix(tf.metrics.Metric):
    """ Computes confusion matrix for mono-label classification. """
    
    @typechecked
    def __init__(self, 
                 num_classes: int,
                 name: str = "Confusion_matrix",
                 dtype = tf.dtypes.int32,
                ):
        super().__init__(name=name, dtype=dtype)
        self.num_classes = num_classes
        self.confusion_matrix = self.add_weight(
            'total',
            shape=[num_classes, num_classes],
            initializer='zeros',
            dtype=self.dtype,
        )

    def update_state(self, y_true, y_pred):
        batch = tf.math.confusion_matrix(
            y_true, y_pred,
            num_classes=self.num_classes,
            dtype=self.dtype,
        )
        self.confusion_matrix.assign_add(batch)

    def result(self):
        return self.confusion_matrix


class MultiLabelConfusionMatrix(tf.metrics.Metric):
    """ Computes confusion matrix for multi-label classification.
    Every class has a dedicated matrix of shape `(2, 2)` 
    that contains:
    - true negatives in cell `(0,0)`
    - false positives in cell `(0,1)`
    - false negatives in cell `(1,0)`
    - true positives in cell `(1,1)`
    Args:
        num_classes: `int`, the number of labels the 
                     prediction task can have.
        name: (Optional) string name of the metric 
              instance.
        dtype: (Optional) data type of the metric 
               result.
    """

    @typechecked
    def __init__(self,
                 num_classes: int,
                 name: str = "Multilabel_confusion_matrix",
                 dtype = tf.dtypes.int32,
        **kwargs,
    ):
        super().__init__(name=name, dtype=dtype)
        self.num_classes = num_classes
        self.true_positives = self.add_weight(
            "true_positives",
            shape=[self.num_classes],
            initializer="zeros",
            dtype=self.dtype,
        )
        self.false_positives = self.add_weight(
            "false_positives",
            shape=[self.num_classes],
            initializer="zeros",
            dtype=self.dtype,
        )
        self.false_negatives = self.add_weight(
            "false_negatives",
            shape=[self.num_classes],
            initializer="zeros",
            dtype=self.dtype,
        )
        self.true_negatives = self.add_weight(
            "true_negatives",
            shape=[self.num_classes],
            initializer="zeros",
            dtype=self.dtype,
        )

    def update_state(self, y_true, y_pred):
        
        y_true = tf.cast(y_true, tf.int32)
        y_pred = tf.cast(y_pred, tf.int32)
        
        true_positive = tf.math.count_nonzero(y_true * y_pred, 0)
        pred_sum = tf.math.count_nonzero(y_pred, 0)
        true_sum = tf.math.count_nonzero(y_true, 0)
        false_positive = pred_sum - true_positive
        false_negative = true_sum - true_positive
        y_true_negative = tf.math.not_equal(y_true, 1)
        y_pred_negative = tf.math.not_equal(y_pred, 1)
        true_negative = tf.math.count_nonzero(
            tf.math.logical_and(y_true_negative, y_pred_negative), axis=0
        )

        self.true_positives.assign_add(tf.cast(true_positive, self.dtype))
        self.false_positives.assign_add(tf.cast(false_positive, self.dtype))
        self.false_negatives.assign_add(tf.cast(false_negative, self.dtype))
        self.true_negatives.assign_add(tf.cast(true_negative, self.dtype))

    def result(self):
        flat_confusion_matrix = tf.convert_to_tensor(
            [
                self.true_negatives,
                self.false_positives,
                self.false_negatives,
                self.true_positives,
            ]
        )
        
        confusion_matrix = tf.reshape(
            tf.transpose(flat_confusion_matrix), [-1, 2, 2]
        )

        return confusion_matrix
    
    def reset_state(self):
        reset_value = np.zeros(self.num_classes, dtype=np.int32)
        tf.keras.backend.batch_set_value([(v, reset_value) for v in self.variables])
    


import tensorflow as tf

def softmax_cross_entropy(labels, logits) -> tf.Tensor:
    """  """
    softmax = tf.nn.softmax_cross_entropy_with_logits
    loss_value = softmax(labels, logits, axis=-1, name='Loss')
    return loss_value



import tensorflow as tf

from transformers import PretrainedConfig
from transformers import TFBertPreTrainedModel

from nlpipes.layers.bert_layers import TFAdaptedBertMainLayer
from nlpipes.data.data_types import TFSequenceClassifierOutput


class TFAdaptedBertForSequenceClassification(TFBertPreTrainedModel):
    """ The Bert Model with a classification head on top.
    It subclass the `TFBertPreTrainedModel` which is the base 
    class for all TensorFlow BERT based models. Note that the 
    `TFBertPreTrainedModel` is also a `tf.keras.Model`.

    Args
    ----------
    bert: 
    dropout: 
    classifier: 
    """
    
    def __init__(self, config: PretrainedConfig, **kwargs):
        super().__init__(config, **kwargs)
        self.bert = TFAdaptedBertMainLayer(config, name="adapted_bert")
        self.dropout= tf.keras.layers.Dropout(config.hidden_dropout_prob)
        self.classifier = tf.keras.layers.Dense(
            units=config.num_labels,
            kernel_initializer=tf.keras.initializers.TruncatedNormal(
                stddev=config.initializer_range),
            name='classifier')
        
    def call(self,
             input_ids,
             attention_mask=None,
             token_type_ids=None,
             position_ids=None,
             head_mask=None,
             inputs_embeds=None,
             output_attentions=None,
             output_hidden_states=None,
             labels=None,
             class_weights=None,
             return_dict=None,
             training = False,
             **kwargs,
            ):
        
        bert_output = self.bert(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            training=training,
        )
        
        pooler_output = bert_output['pooler_output']
        dropout_output = self.dropout(pooler_output, training=training)
        logits = self.classifier(dropout_output)
        
        return TFSequenceClassifierOutput(
            logits=logits,
            hidden_states=bert_output.hidden_states,
            attentions=bert_output.attentions,
    )


class AdaptedBertForSequenceClassification(TFBertPreTrainedModel):
    """ The Bert Model with a classification head on top.
    It subclass the `TFBertPreTrainedModel` which is the base 
    class for all TensorFlow BERT based models. Note that the 
    `TFBertPreTrainedModel` is also a `tf.keras.Model`.

    Args
    ----------
    bert: 
    dropout: 
    classifier: 
    """
    
    def __init__(self, config: PretrainedConfig, **kwargs):
        super().__init__(config, **kwargs)
        self.bert = TFAdaptedBertMainLayer(config, name="adapted_bert")
        self.dropout= tf.keras.layers.Dropout(config.hidden_dropout_prob)
        self.classifier = tf.keras.layers.Dense(
            units=config.num_labels,
            kernel_initializer=tf.keras.initializers.TruncatedNormal(
                stddev=config.initializer_range),
            name='classifier')
        
    def call(self,
             input_ids,
             attention_mask=None,
             token_type_ids=None,
             position_ids=None,
             head_mask=None,
             inputs_embeds=None,
             output_attentions=None,
             output_hidden_states=None,
             labels=None,
             class_weights=None,
             return_dict=None,
             training = False,
             **kwargs,
            ):
        
        bert_output = self.bert(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            training=training,
        )
        
        pooler_output = bert_output['pooler_output']
        dropout_output = self.dropout(pooler_output, training=training)
        logits = self.classifier(dropout_output)
        
        return TFSequenceClassifierOutput(
            logits=logits,
            hidden_states=bert_output.hidden_states,
            attentions=bert_output.attentions,
    )
    


from typing import (
    List,
    Optional,
)

from dataclasses import dataclass

import tensorflow as tf
import transformers

from transformers import (
    PretrainedConfig,
    PreTrainedTokenizerFast,
    TFPreTrainedModel,
)

from tokenizers.pre_tokenizers import Whitespace

from nlpipes.losses.losses import softmax_cross_entropy

from nlpipes.data.data_selectors import DataSelector
from nlpipes.data.data_augmentors import VocabAugmentor
from nlpipes.data.data_processors import DataProcessorForSequenceClassification
from nlpipes.data.data_cleaners import clean
from nlpipes.data.data_utils import chunk

from nlpipes.data.data_types import (
     Corpus,
     Document,
     Token,
     InputExample,
     InputTokens,
     InputFeatures,
     Predictions,
     Outcomes,
)

def select_corpus(
    corpus: Corpus,
    tokenizer: PreTrainedTokenizerFast,
    target_corpus_size: int,
    similarity_metrics: List[str],
    diversity_metrics: List[str],
) -> Corpus :
    """ """

    data_selector = DataSelector(
        name='DataSelection',
        tokenizer=tokenizer,
        target_corpus_size=target_corpus_size,
        similarity_metrics=similarity_metrics,
        diversity_metrics=diversity_metrics,
    )

    clean_corpus = [clean(text) for text in corpus]
    sub_corpus= data_selector(clean_corpus)

    return sub_corpus

def augment_vocabulary(
    corpus: Corpus,
    tokenizer: PreTrainedTokenizerFast, 
    target_vocab_size: int,
) -> List[Token]:
    """ """

    vocab_augmentor = VocabAugmentor(
        name='VocabAugmentation',
        tokenizer=tokenizer,
        target_vocab_size=target_vocab_size,
        cased=False, 
    )

    new_tokens = vocab_augmentor.get_new_tokens(corpus)

    return new_tokens

def preprocess(texts: Corpus,
               labels: Optional[str] = None,
               chunk_text: Optional[bool] = False,
              ) -> List[InputExample]:
    """ Serialize each raw text into examples. Raw texts can 
    eventually be chunked into spans. E.g., at inference time,
    each span need to be chunked into sentences if the model 
    has been trained at the sentence level. """

    if labels is not None:
        input_examples = [
            InputExample(text=text, label=label)
            for text, label in zip(texts, labels)
        ]
    else:
        spans = chunk(text) if chunk_text == True else texts
        input_examples = [
            InputExample(span) for span in spans
        ]

    return input_examples

def tokenize(examples: List[InputExample],
             tokenizer: PreTrainedTokenizerFast,
            ) -> InputTokens:
    """ Convert examples into tokens. Tokens encapsulate both 
    words, wordpieces and enriched wordpieces with the special
    [CLS] and [SEP] tokens. """

    texts = [example.text for example in examples]

    words = [Whitespace().pre_tokenize_str(text) for text in texts]

    wordpieces = [
        encoding.tokens for encoding 
        in tokenizer.backend_tokenizer.encode_batch(
           texts, add_special_tokens=False
        )
    ]

    tokens = [
        encoding.tokens for encoding 
        in tokenizer.backend_tokenizer.encode_batch(
           texts, add_special_tokens=True
        )
    ]

    input_tokens = InputTokens(
        words=words,
        wordpieces=wordpieces,
        tokens=tokens,
    )

    return input_tokens

def encode(tokens: InputTokens,
           tokenizer: PreTrainedTokenizerFast,
          ) -> InputFeatures: 
    """ Encode human-readable wordpieces into encoded features
    that can be use as input by the model. """

    inputs = tokenizer(
       tokens.wordpieces,
       add_special_tokens=False,
       padding=True,
       truncation=True,
       max_length=512,
       is_split_into_words = True,
       return_tensors='tf',
    )

    input_features = InputFeatures(
       input_ids=inputs['input_ids'],
       attention_mask=inputs['attention_mask'],
       token_type_ids=inputs['token_type_ids'],
    )

    return input_features

def estimate(features: InputFeatures,
             model: TFPreTrainedModel,
             config: PretrainedConfig,
            ) -> Predictions:
    """ Predict the label index matching the maximum prediction
    score and compute the loss value associated with each
    prediction."""

    outputs = model.call(
        input_ids=features.input_ids,
        attention_mask=features.attention_mask,
        token_type_ids=features.token_type_ids,
    )

    logits = outputs.logits
    label = tf.argmax(logits, axis=-1)
    label_onehot = tf.one_hot(label, config.num_labels)
    scores = tf.nn.softmax(logits, axis=1)
    losses = softmax_cross_entropy(label_onehot, logits)

    predictions = Predictions(
        label=label,
        scores=scores,
        losses=losses,
    )

    return predictions

def postprocess(predictions: Predictions, 
                config: PretrainedConfig,
               ) -> Outcomes:
    """ Convert outputs tensors into more handy numpy lists. """

    indice = [label.numpy() for label in predictions.label]
    scores = [
        score.numpy().tolist() 
        for score in predictions.scores
    ] 
    losses = [loss.numpy() for loss in predictions.losses]

    label = [config.id2label[idx] for idx in indice]

    outcomes = Outcomes(
        indice=indice,
        label=label,
        scores=scores,
        losses=losses,
    )

    return outcomes





import tensorflow as tf

from typeguard import typechecked


class GradientAccumulator(tf.optimizers.Optimizer):
    """Optimizer wrapper for gradient accumulation."""
    @typechecked
    def __init__(
        self,
        optimizer: tf.optimizers,
        accum_steps: int = 4,
        name: str = "GradientAccumulator",
        **kwargs,
    ):
        """ Construct a GradientAccumulator optimizer."""
        super().__init__(name, **kwargs)
        self._optimizer = tf.keras.optimizers.get(optimizer)
        self._gradients = []
        self._accum_steps = accum_steps

    def _create_slots(self, var_list):
        self._optimizer._create_slots(var_list=var_list)
        for var in var_list:
            self.add_slot(var, "ga")

        self._gradients = [self.get_slot(var, "ga") for var in var_list]

    @property
    def gradients(self):
        """The accumulated gradients on the current replica."""
        if not self._gradients:
            raise ValueError(
                "The accumulator should be called first to initialize the gradients"
            )
        return list(
            gradient.read_value() if gradient is not None else gradient
            for gradient in self._gradients
        )

    def apply_gradients(self, grads_and_vars, name=None, **kwargs):
        self._optimizer._iterations = self.iterations
        return super().apply_gradients(grads_and_vars, name, **kwargs)

    def _resource_apply_dense(self, grad, var, apply_state=None):
        accum_gradient = self.get_slot(var, "ga")
        if accum_gradient is not None and grad is not None:
            accum_gradient.assign_add(
                grad, use_locking=self._use_locking, read_value=False
            )

        def _apply():
            if "apply_state" in self._optimizer._dense_apply_args:
                train_op = self._optimizer._resource_apply_dense(
                    accum_gradient.read_value(), var, apply_state=apply_state
                )
            else:
                train_op = self._optimizer._resource_apply_dense(
                    accum_gradient.read_value(), var
                )
            reset_op = accum_gradient.assign(
                        tf.zeros_like(accum_gradient),
                        use_locking=self._use_locking,
                        read_value=False,
                    )
            return tf.group(train_op, reset_op)

        apply_op = tf.cond(
            (self.iterations+1) % self._accum_steps == 0, _apply, lambda: tf.no_op()
        )
        return apply_op

    def _resource_apply_sparse(self, grad: tf.Tensor, var, indices, apply_state):
        accum_gradient = self.get_slot(var, "ga")
        if accum_gradient is not None and grad is not None:
            self._resource_scatter_add(accum_gradient, indices, grad)

        def _apply():
            if "apply_state" in self._optimizer._sparse_apply_args:
                train_op = self._optimizer._resource_apply_sparse(
                    accum_gradient.sparse_read(indices),
                    var,
                    indices,
                    apply_state=apply_state,
                )
            else:
                train_op = self._optimizer._resource_apply_sparse(
                    accum_gradient.sparse_read(indices), var, indices
                )
            reset_op = accum_gradient.assign(
                        tf.zeros_like(accum_gradient),
                        use_locking=self._use_locking,
                        read_value=False,
                    )
            return tf.group(train_op, reset_op)

        apply_op = tf.cond(
            (self.iterations+1) % self._accum_steps == 0, _apply, lambda: tf.no_op()
        )
        return apply_op

    def reset(self):
        """Resets the accumulated gradients on the current replica."""
        assign_ops = []
        if not self._gradients:
            return assign_ops

        for gradient in self._gradients:
            if gradient is not None:
                assign_ops.append(
                    gradient.assign(
                        tf.zeros_like(gradient),
                        use_locking=self._use_locking,
                        read_value=False,
                    )
                )

        return tf.group(assign_ops)

    @property
    def lr(self):
        return self._optimizer._get_hyper("learning_rate")

    @lr.setter
    def lr(self, lr):
        self._optimizer._set_hyper("learning_rate", lr)  #

    @property
    def learning_rate(self):
        return self._optimizer._get_hyper("learning_rate")

    @learning_rate.setter
    def learning_rate(self, learning_rate):
        self._optimizer._set_hyper("learning_rate", learning_rate)

    def get_config(self):
        config = {"accum_steps": self._accum_steps}
        base_config = super().get_config()
        return {**base_config, **config}
    


import os
import importlib
import logging
import inspect

from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
)

import tensorflow as tf
import transformers

from transformers import (
    TFPreTrainedModel,
    TFAutoModelForMaskedLM,
    TFAutoModelForSequenceClassification,
    TFBertForSequenceClassification,
    BertTokenizerFast,
    AutoModel,
    AutoTokenizer,
    AutoConfig, 
)

from nlpipes.data.data_types import Task
from nlpipes.pipelines.base import TFPipeline
from nlpipes.pipelines.language_modeling import TFPipelineForMaskedLM
from nlpipes.pipelines.sequence_classification import TFPipelineForSequenceClassification
from nlpipes.models.sequence_classification import TFAdaptedBertForSequenceClassification
from nlpipes.models.sequence_classification import AdaptedBertForSequenceClassification
from nlpipes.configurations.bert_config import AdaptedBertConfig
from nlpipes.data.data_types import LabelsMapping

tf.get_logger().setLevel('ERROR')

logger = logging.getLogger('__name__')


SUPPORTED_TASKS = {
  "domain-adaptation": {
      "pipeline": TFPipelineForMaskedLM,
      "models": {
          "default": TFAutoModelForMaskedLM,
          "bottleneck-adapters": None,
      },
      "tokenizers": {
          "default": AutoTokenizer,
          "bottleneck-adapters": AutoTokenizer,
      },
      "configs": {
          "default": AutoConfig,
          "bottleneck-adapters": AdaptedBertConfig,
      },
  },
  "single-label-classification": {
      "pipeline": TFPipelineForSequenceClassification,
      "models": {
          "default": TFAutoModelForSequenceClassification,
          "bottleneck-adapters": TFAdaptedBertForSequenceClassification,
      },
      "tokenizers": {
          "default": AutoTokenizer,
          "bottleneck-adapters": AutoTokenizer,
      },
      "configs": {
          "default": AutoConfig,
          "bottleneck-adapters": AdaptedBertConfig,
      },
  },
}


def Model(
    name_or_path: TFPreTrainedModel = None,
    task: str = None,
    all_labels: Optional[List] = None,
    **model_kwargs
) -> TFPipeline:

    """ The `Model` is an abstraction that return a pipeline.
    Which pipeline the model shall call is determined
    automatically from the target task supplied to train the
    model. For more information, see the task argument.

    Args
    ----------
    task(str):
          Determines which task the model should learn. Setting
          the task manually will be not needed in the future as it
          will be determined automatically from the input data. But
          it has been implemented for more flexibility in future 
          developments. Available tasks currently are :
            - `mono_label` classification`
            - `class_label` classification`
            - `multi_label` classification`
            
    model(str):
          A path or name of the model that should be used.
          It can be:
            - The path of a model available in a local repo.
            - The name of a model available in the Huggingface
              model repo.
    
    all_labels (List[str]): 
          A list of all labels. Only used for classification task.
    """

    try:
        pipeline = create_pipeline_from_task(
            task=task, 
            name_or_path=name_or_path, 
            all_labels=all_labels,
            customization_tag='#',
        )

        return pipeline

    except EnvironmentError as error:
        message = "The pipeline, model, tokenizer or config file could not be found."
        logger.error(message)
        raise error


def create_pipeline_from_task(
    task:str, 
    name_or_path:str,
    all_labels:List[str],
    customization_tag=str,
) -> TFPipeline:
    """ Create the pipeline given a downstream task. """
    
    if os.path.exists(name_or_path):
        config = AutoConfig.from_pretrained(name_or_path)
        PIPELINE = SUPPORTED_TASKS[config.finetuning_task]['pipeline']
        
        MODEL = get_model_from_config(config)
        
        model = MODEL.from_pretrained(name_or_path)
        tokenizer = AutoTokenizer.from_pretrained(name_or_path)
        
        pipeline = PIPELINE(
            model=model,
            tokenizer=tokenizer,
            config=config,
        )

    else:
        TASK = SUPPORTED_TASKS[task]
        PIPELINE = TASK["pipeline"]
    
        if customization_tag in name_or_path:
            customization = name_or_path.split(customization_tag, 1)[1]
            MODEL = TASK["models"][customization]
            TOKENIZER = TASK["tokenizers"][customization]
            CONFIG = TASK["configs"][customization]
            name_or_path = name_or_path.replace(
                customization_tag+customization, ""
            )
        else:
            MODEL = TASK["models"]["default"]
            TOKENIZER = TASK["tokenizers"]["default"]
            CONFIG = TASK["configs"]["default"]

        config = CONFIG.from_pretrained(name_or_path)
        
        if all_labels:
            config.id2label=LabelsMapping(all_labels).id2label()
            config.label2id=LabelsMapping(all_labels).label2id()
            config.num_labels=LabelsMapping(all_labels).num_labels()
            
        if not config.finetuning_task:
            config.finetuning_task=task

        tokenizer = TOKENIZER.from_pretrained(name_or_path)
        model = MODEL.from_pretrained(name_or_path, config=config)
        
        pipeline = PIPELINE(
            model=model,
            tokenizer=tokenizer,
            config=config,
        )
    
    return pipeline


def get_model_from_config(config):
    """ Get the model class from the model configuration file.
    Note : we use a temporary dirty workaround to bypass an issue in 
    the model architecture name stored in the saved config """
    
    model_name = config.architectures[0]
    models_module_name = 'nlpipes.models.sequence_classification'
    models_module = importlib.import_module(models_module_name)
    available_models = [
        model[0] for model in inspect.getmembers(models_module)
    ]
    
    if model_name in available_models:
        model_class = getattr(models_module, model_name)
    else:
        model_class = getattr(transformers, 'TF' + model_name)
    return model_class


def get_available_tasks() -> List[str]:
    """ Get a list of all available tasks."""
    return list(SUPPORTED_TASKS.keys())



import os
import sys
import logging
import time
import datetime

from abc import ABC, abstractmethod
from collections import defaultdict

from dataclasses import astuple
from dataclasses import dataclass
from dataclasses import field

import numpy as np
import tensorflow as tf
from transformers import TFPreTrainedModel

from tqdm.auto import tqdm

from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
)

from nlpipes.data.data_types import (
    InputFeatures,
    TFSequenceClassifierBatchOutput,
)

logger = logging.getLogger('__name__')


"""
obj:`Callbacks` encapsulate a set of functions to be applied at given
stages of the training procedure to customize the trainer behavior. 
The relevant callbacks functions will then be called at the 
various stages of the model training lifecycle, such as when:
 - the training starts.
 - the training ends.
 - the testing starts.
 - the testing ends.
 - an epoch starts
 - an epoch ends.
 - a batch starts.
 - a batch ends.
"""

class Callback(ABC):
    
    """ Abstract base class for callback objects. """
    
    def on_train_begin(self):
        """ Called at the beginning of the training. """
     
    def on_train_end(self):
        """ Called at the end of the training. """
      
    def on_test_begin(self):
        """ Called at the beginning of the evaluation. """
     
    def on_test_end(self):
        """ Called at the end of the evaluation. """
    
    def on_epoch_begin(self, *args):
        """ Called at the beginning of an epoch. """

    def on_epoch_end(self, *args):
        """ Called at the end of an epoch. """
    
    def on_train_batch_begin(self, *args):
        """ Called at the beginning of a training batch. """
    
    def on_train_batch_end(self, *args):
        """ Called at the end of a training batch. """
    
    def on_test_batch_begin(self, *args):
        """ Called at the beginning of a test batch. """

    def on_test_batch_end(self, *args):
        """ Called at the end of a testing batch. """


@dataclass
class CallbackList(Callback):
        
    """ Internal class that just calls the list of callbacks.
    The CallbackList serves two main purposes:
        1. It stores all the callbacks.
        2. It allows to call all of the individual callbacks 
           easily. For example, if we have three callbacks 
           that do something at the end of an epoch, then 
           callbacks.on_epoch_end() will call the three
           callback objects in order. 
    """

    callbacks: List[Callback]
   
    def on_train_begin(self, *args):
        for callback in self.callbacks:
            callback.on_train_begin(*args)
        
    def on_train_end(self, *args):
        for callback in self.callbacks:
            callback.on_train_end(*args)
        
    def on_test_begin(self, *args):
        for callback in self.callbacks:
            callback.on_test_begin(*args)
        
    def on_test_end(self, *args):
        for callback in self.callbacks:
            callback.on_test_end(*args)

    def on_epoch_begin(self, *args):
        for callback in self.callbacks:
            callback.on_epoch_begin(*args)

    def on_epoch_end(self, *args):
        for callback in self.callbacks:
            callback.on_epoch_end(*args)
            
    def on_train_batch_begin(self, *args):
        for callback in self.callbacks:
            callback.on_train_batch_begin(*args)

    def on_train_batch_end(self, *args):
        for callback in self.callbacks:
            callback.on_train_batch_end(*args)

    def on_test_batch_begin(self, *args):
        for callback in self.callbacks:
            callback.on_test_batch_begin(*args)

    def on_test_batch_end(self, *args):
        for callback in self.callbacks:
            callback.on_test_batch_end(*args) 


@dataclass
class TrainingStep(Callback):
    """ Callback that define the training step. The training step is
    a single iteration of the training process. It consists of a 
    forward pass (where the obj:`model` makes predictions based on
    the input data) and a backward pass (where the obj:`optimizer`
    updates the obj:`model` weights and biases based on the difference
    between the predicted output and the expected output as expressed
    by the obj:`loss_finction`).
    """
    
    name: str = 'TrainingStep'
    model: TFPreTrainedModel = None
    optimizer: tf.optimizers.Optimizer = None
    loss_function: tf.losses = None
    
    def _train_step(self, features: InputFeatures):
        """ Train the obj:`model` on a single batch of examples. """
        
        with tf.GradientTape() as tape:
            model_outputs = self.model.call(
                input_ids=features.input_ids,
                attention_mask=features.attention_mask,
                token_type_ids=features.token_type_ids,
                training=True,
            )
            
            loss_values = self.loss_function(
                features.label, model_outputs.logits
            )
            
        trainable_variables = self.model.trainable_variables
            
        gradients = tape.gradient(loss_values, trainable_variables)
        
        self.optimizer.apply_gradients(
            zip(gradients, trainable_variables)
        )
        
        output_batch = TFSequenceClassifierBatchOutput(
            loss=loss_values,
            logits=model_outputs.logits,
            hidden_states=model_outputs.hidden_states,
            attentions=model_outputs.attentions,
        )
        
        return output_batch
    
    def _test_step(self, features: InputFeatures):
        """ Test the obj:`model` on a single batch of examples. """
        
        model_outputs = self.model.call(
                  input_ids=features.input_ids,
                  attention_mask=features.attention_mask,
                  token_type_ids=features.token_type_ids,
                  training=False,
        )
        
        loss_values = self.loss_function(
            features.label, model_outputs.logits
        )
        
        return TFSequenceClassifierBatchOutput(
            loss=loss_values,
            logits=model_outputs.logits,
            hidden_states=model_outputs.hidden_states,
            attentions=model_outputs.attentions,
        )
    
    def on_train_begin(self):
        message = self.model.summary()
        logger.info(message)           
     
    def on_train_batch_begin(self, step: int, features: InputFeatures):
        self.train_step_output = self._train_step(features)
             
    def on_test_batch_begin(self, step: int, features: InputFeatures):
        self.test_step_output = self._test_step(features)


@dataclass
class History(Callback):
    """ Callback that records the performance of the model 
    during training. It stores all the evaluation metrics 
    such as loss and accuracy at each training and testing
    steps. This allows to track the progress of the model
    over time, and can be used by other callbacks, e.g. to
    detect when the model has reached its optimal performance.
    """
    
    name: str = 'History'
    epoch: int = field(default_factory=int)
    metric: Callable[[], tf.metrics.Metric] = None
    training_step: Callback = None
    train: Dict = field(default_factory=lambda: defaultdict(dict))
    test: Dict = field(default_factory=lambda: defaultdict(dict))
    train_details: Dict = field(default_factory=lambda: defaultdict(dict))
    test_details: Dict = field(default_factory=lambda: defaultdict(dict))

    
    def __post_init__(self):
        self.train_metric = self.metric()
        self.test_metric = self.metric()
    
    def on_epoch_begin(self, epoch: int):
        """ Resets all of the metric state variables """
        self.epoch = epoch
        self.train_metric.reset_states()
        self.test_metric.reset_states()
        
        self.train['loss'] = {}
        self.train_details['loss'] = {}
        self.test_details['loss'] = {}
        self.train['loss'][epoch] = []
        self.train_details['loss'][epoch] = []
        self.test_details['loss'][epoch] = []
        
        self.test['accuracy'] = {}
        self.train_details['accuracy'] = {}
        self.test_details['accuracy'] = {}
        self.test['accuracy'][epoch] = []
        self.train_details['accuracy'][epoch] = []
        self.test_details['accuracy'][epoch] = []
    
    def on_epoch_end(self, epoch: int):
        """ Display the performance metrics after each epoch """
        self.train['loss'][self.epoch] = self.train_metric.result()
        self.test['loss'][self.epoch] = self.test_metric.result()
        
        message = f'Epoch {self.epoch+1:4d} {self.name:10} ' \
                  f'Train {self.train_metric.result():1.3f} ' \
                  f'Test {self.test_metric.result():1.3f}'
        
        logger.info(message)

    def on_train_batch_end(self, step: int):
        """ Compute the performance metrics after each training batch """
        loss = self.training_step.train_step_output.loss
        self.train_metric(loss)
        self.train_details['loss'][self.epoch].extend(loss)
        self.train['loss'][self.epoch] = self.train_metric.result()
        
    def on_test_batch_end(self, step: int):
        """ Compute the performance metrics after each testing batch """
        loss = self.training_step.test_step_output.loss
        self.test_metric(loss)
        self.test_details['loss'][self.epoch].extend(loss)
        self.test['loss'][self.epoch] = self.test_metric.result()
        
    @property
    def best_result(self) -> float:
        return min(self.test['loss'].values())
    

@dataclass
class ModelCheckpoint(Callback):
    """ Callback that save the model weights and configuration 
    after each epoch and to keep track of the best model """
    
    name: str = 'ModelCheckpointing'
    model: TFPreTrainedModel = None
    history: History = None
    direction: str = 'minimize'
    best_result: float = np.inf
    best_model_dir: str = ''
    checkpoints_dir: str = 'checkpoints'

    def __post_init__(self):
        """ Create the directory for saving checkpoints. """
        if self.direction not in ['minimize', 'maximize']:
            raise ValueError
        if self.direction == 'maximize':
            self.best_result = 0

        if not os.path.isdir(self.checkpoints_dir):
            abs_path = os.path.abspath(self.checkpoints_dir)
            text = f'Make a checkpoint directory: {abs_path}'
            logger.info(text)
            os.makedirs(self.checkpoints_dir)

    def on_epoch_end(self, epoch: int):
        """ Pass the obj:`ModelCheckpoint` callback after the obj:`History`. """
        result = self.history.test['loss'][epoch]
        diff = self.best_result - result
        is_better = diff > 0 if self.direction == 'minimize' else diff < 0
        if is_better:
            name = f'epoch-{epoch:02d}-{result:.2f}'
            model_dir = os.path.join(self.checkpoints_dir, name)
            os.mkdir(model_dir)
            self.model.config.architectures = self.model.__class__.__name__
            self.model.save_pretrained(model_dir)
            self.best_result = result
            self.best_model_dir = model_dir
            message = f'New model checkpoint saved'
            logger.info(message)


@dataclass
class CSVLogger(Callback):
    """ Callback that save the logs of the performance metrics
    into a csv file. """
    
    name: str = 'CSVLogging'
    logs_dir: str = None
    level: int = 20
    msg_format: str = '%(asctime)s [%(levelname)-5s] %(message)s'

    def __post_init__(self):
        root_logger = logging.getLogger('__name__')
        root_logger.setLevel(self.level)
        root_logger.propagate = False
        formatter = logging.Formatter(self.msg_format, 
                                      datefmt='%Y-%m-%d %H:%M:%S')
        console = logging.StreamHandler()
        console.setFormatter(formatter)
        
        root_logger.addHandler(console)
        if self.logs_dir:
            logs_dir = os.path.abspath(self.logs_dir)
            os.makedirs(logs_dir, exist_ok=True)
            file_path = os.path.join(logs_dir, 'experiment.log')
            file_handler = logging.FileHandler(file_path, mode='w')
            file_handler.setFormatter(formatter)
            root_logger.addHandler(file_handler)


@dataclass
class EarlyStopping(Callback):
    """ Callback that stop training the model once a certain metric
    is no longer improving. The metric used is typically the 
    validation loss, but it can also be a metric such as accuracy
    or precision """
    
    name: str = 'EarlyStopping'
    history: History = None
    patience: int = 3
    min_delta: float = 0.01
    best_result: float = np.inf
    current_patience: int = 0
    direction: str = 'maximize'

    def __post_init__(self):
        """ """
        if self.direction not in ['minimize', 'maximize']:
            raise ValueError
        if self.direction == 'maximize':
            self.best_result = 0

    def on_epoch_end(self, epoch: int):
        """ """
        result = self.history.test['loss'][epoch]
        diff = self.best_result - result
        is_better = diff > self.min_delta if self.direction == 'minimize' \
            else diff < self.min_delta * -1
        if is_better:
            self.best_result = result
            self.current_patience = 0
            return
        self.current_patience += 1
        if self.current_patience >= self.patience:
            raise Exception('Early stopping of the training')


@dataclass
class TimedStopping(Callback):
    """ Callbacks that stop training the model when a specified
    amount of time has passed.
    Args:
        time_limit: maximum amount of time in seconds before stopping.
        stopped_epoch: the epoch number where the training loop stopped.
    """

    name: str = 'TimedStopping'
    time_limit: int = 86000
    stopped_epoch: int = None

    def get_config(self):
        config = {"seconds": self.time_limit}
        base_config = super().get_config()
        return {**base_config, **config}
    
    def on_train_begin(self):
        self.stopping_time = time.time() + self.time_limit
        
    def on_epoch_end(self, epoch: int):
        if time.time() >= self.stopping_time:
            self.stopped_epoch = epoch
            formatted_time = datetime.timedelta(seconds=self.time_limit)
            message = "Timed stopping at epoch {} after training for {}".format(
                self.stopped_epoch + 1, formatted_time
            )
            raise Exception(message)

    def on_train_end(self):
        if self.stopped_epoch is not None:
            formatted_time = datetime.timedelta(seconds=self.time_limit)
            message = "Timed stopping at epoch {} after training for {}".format(
                self.stopped_epoch + 1, formatted_time
            )
            logger.info(message)


@dataclass
class ProgbarLogger(Callback):
    """ Callbacks to monitor the progress of a model's training, 
    and display the progress bar and performance metrics in the
    console. """ 
    
    name: str = 'ProgbarLogger'
    history: List[History] = None
    metric: Callable = tf.metrics.Mean
    epoch: int = field(default_factory=int)
    num_samples: int = None
    num_epochs: int = None
    batch_size: int = None
    show_overall_progress: bool = False
    show_epoch_progress: bool = True
    update_per_second: int = 10
    is_training = False

    def __post_init__(self):
        self.tqdm = tqdm
        self.num_steps: int = int(self.num_samples/self.batch_size)
        self.update_interval = 1/self.update_per_second
        self.last_update_time = time.time()
        self.overall_bar_format: str = "{l_bar}{bar} {n_fmt}/{total_fmt} " \
                                       "ETA: {remaining}s, {rate_fmt}{postfix}"
        self.epoch_bar_format: str = "Batch {n_fmt}/{total_fmt}{bar} " \
                                     "ETA: {remaining}s - {desc}"
        self.metrics_format: str = "{name}: {value:0.4f}" 
        self.metrics_separator: str = " - "
        
    def _initialize_progbar(self, hook, epoch):
        self.num_samples_seen = 0
        self.steps_to_update = 0
        self.steps_so_far = 0
        self.train_logs = defaultdict(float)
        self.test_logs = defaultdict(float)
        
        if hook == "train":
            current_epoch_description = "Epoch {epoch}/{num_epochs}".format(
                epoch=epoch+1, num_epochs=self.num_epochs
            )
            if self.show_epoch_progress:
                print(current_epoch_description)
                self.epoch_progress_tqdm=self.tqdm(
                    total=self.num_steps,
                    position=0,
                    bar_format=self.epoch_bar_format,
                    leave=False,
                    dynamic_ncols=True,
                    unit="steps",
                )
        elif hook == "test":
            if self.show_epoch_progress:
                self.epoch_progress_tqdm=self.tqdm(
                    desc="Testing progress",
                    total=self.num_steps,
                    position=0,
                    bar_format=self.epoch_bar_format,
                    leave=False,
                    dynamic_ncols=True,
                    unit="steps",
                )

    def _update_progbar(self):
        self.num_samples_seen += self.batch_size
        self.steps_to_update += 1
        self.steps_so_far += 1

        if self.steps_so_far <= self.num_steps:  
            now = time.time()
            time_diff = now - self.last_update_time
            if self.show_epoch_progress and time_diff >= self.update_interval:
                
                self.epoch_progress_tqdm.desc = \
                     f'loss: {self.train_logs["loss"]:04F} - ' \
                     f'val_loss: {self.test_logs["loss"]:04F}'
                     #f'acc: {self.logs["accuracy"]:04F} - ' \
                     #f'val_acc: {self.logs["test_accuracy"]:04F}'

                self.epoch_progress_tqdm.update(self.steps_to_update)
                self.steps_to_update = 0
                self.last_update_time = now
                
    def _clean_up_progbar(self, hook):
        if self.show_epoch_progress:
            self.epoch_progress_tqdm.miniters = 0
            self.epoch_progress_tqdm.mininterval = 0
            self.epoch_progress_tqdm.update(
               self.num_steps - self.epoch_progress_tqdm.n
           )
            self.epoch_progress_tqdm.close()
                 
    def on_train_begin(self):
        self.is_training = True
        
    def on_train_end(self):
        self.is_training = False
        self._clean_up_progbar("train")
        
    def on_test_begin(self):
        if not self.is_training:
            self._initialize_progbar("test", epoch=None)
            
    def on_test_end(self):
        if not self.is_training:
            self._clean_up_progbar("test")
     
    def on_epoch_begin(self, epoch: int):
        self.epoch = epoch
        self._initialize_progbar("train", epoch)
    
    def on_epoch_end(self, epoch: int):
        self._clean_up_progbar("train")
        if self.show_overall_progress:
            self.overall_progress_tqdm.update(1)
        self.epoch_progress_tqdm.close()

    def on_train_batch_end(self, step: int):
        train_logs = defaultdict(float)
        train_logs["loss"] = self.history.train['loss'][self.epoch]
        self.train_logs = train_logs
        self._update_progbar()
        
    def on_test_batch_end(self, step: int):
        test_logs = defaultdict(float)
        test_logs["loss"] = self.history.test['loss'][self.epoch]
        self.test_logs = test_logs 
        self._update_progbar()



import logging

from typing import Iterable
from typing import List

from dataclasses import dataclass

from nlpipes.callbacks.callbacks import Callback
from nlpipes.callbacks.callbacks import CallbackList
from nlpipes.data.data_types import InputFeatures


logger = logging.getLogger('__name__')

@dataclass
class Trainer():
    
    """ The trainer is written from scratch. It consists in 
    a custom training loop that trigger a set of `Callback`
    functions at given stages of the training procedure, 
    such as when:
     - the training/testing starts.
     - the training/testing ends.
     - a train/test epoch starts
     - a train/test epoch ends.
     - a train/test batch starts.
     - a train/testbatch ends. 
    """
    
    train_stream: Iterable[InputFeatures]
    test_stream: Iterable[InputFeatures]
    callbacks: List[Callback]
    num_epochs: int
    
    def train(self):
        
        callbacks = CallbackList(self.callbacks)
        
        try:
            callbacks.on_train_begin()
            
            for epoch in range(self.num_epochs):
                
                callbacks.on_epoch_begin(epoch)
                
                for step, batch in enumerate(self.train_stream):
                    callbacks.on_train_batch_begin(step, batch)
                    callbacks.on_train_batch_end(step)
                    
                if self.test_stream:
                    for step, batch in enumerate(self.test_stream):
                        callbacks.on_test_batch_begin(step, batch)
                        callbacks.on_test_batch_end(step)
                       
                callbacks.on_epoch_end(epoch)
                
            callbacks.on_train_end()
             
        except:
            message = 'Unexpected error during the training process.'
            logger.error(message)
            exit()
            


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nlpipes',
 'nlpipes..ipynb_checkpoints',
 'nlpipes.callbacks',
 'nlpipes.callbacks..ipynb_checkpoints',
 'nlpipes.configurations',
 'nlpipes.data',
 'nlpipes.data..ipynb_checkpoints',
 'nlpipes.layers',
 'nlpipes.losses',
 'nlpipes.losses..ipynb_checkpoints',
 'nlpipes.metrics',
 'nlpipes.metrics..ipynb_checkpoints',
 'nlpipes.models',
 'nlpipes.models..ipynb_checkpoints',
 'nlpipes.optimization',
 'nlpipes.optimization..ipynb_checkpoints',
 'nlpipes.pipelines',
 'nlpipes.pipelines..ipynb_checkpoints',
 'nlpipes.trainers',
 'nlpipes.trainers..ipynb_checkpoints']

package_data = \
{'': ['*']}

install_requires = \
['ftfy>=6.1.1,<7.0.0',
 'numpy>=1.21.0,<2.0.0',
 'pandas>=1.3.5,<2.0.0',
 'scipy>=1.7.3,<2.0.0',
 'tensorflow>=2.11.0,<3.0.0',
 'tokenizers>=0.13.0,<0.14.0',
 'tqdm>=4.64.0,<5.0.0',
 'transformers>=4.24.0,<5.0.0']

setup_kwargs = {
    'name': 'nlpipes',
    'version': '0.1.7',
    'description': 'Text Classification with Transformers',
    'long_description': '<!-- PROJECT NAME -->\n<div align="center">\n  <h2>NLPIPES</h2>\n  <h3>Text Classification with Transformers</h3>\n</div>\n\n<div align="center">\n    <a href="https://opensource.org/licenses/Apache-2.0">\n       <img alt="Licence" src="https://img.shields.io/badge/License-Apache_2.0-blue.svg">\n    </a>\n     <a href="https://opensource.org/licenses/Apache-2.0">\n       <img alt="PyPi Version" src="https://img.shields.io/pypi/pyversions/nlpipes">\n    </a> \n    <a href="https://pypi.org/project/nlpipes/">\n        <img alt="PyPi Package Version" src="https://img.shields.io/pypi/v/nlpipes">\n    </a>\n    <!--\n    <a href="https://pepy.tech/project/nlpipes/">\n        <img alt="PyPi Downloads" src="https://static.pepy.tech/badge/nlpipes/month">\n    </a>\n    -->\n</div>\n\n<div align="center">\n    <a href=""><strong>Documentation</strong></a>\n    • <a href=""><strong>References</strong></a>\n</div>\n\n<div align="center">\n  <img src="https://raw.githubusercontent.com/uyanik/Images/dd6d7682a6b0e1bc504a5b9da1fe930e01a39aab/nlpipes_screenshot.png" alt="nlpipes_screenshot" title="nlpipes screenshot">\n</div>\n\n`NLPipes` provides an easy way to use Transformers-based models for training, evaluation and inference on a diversity of text classification tasks, including:\n\n* **Single-label classification**: Assign one label to each text. A typical use case is sentiment analysis where one want to detect the overall sentiment polarity (e.g., positive, neutral, negative) in a review.\n* **Multi-labels classification** [Not yet implemented]: Assign one or more label to each text from a list of possible labels. A typical use case is categories detection where one want to detect the multiple aspects mentionned in a review (e.g., #product_quality, #delivery_time, #price, ...).\n* **Aspect-based classification** [Not yet implemented]: Assign one label from a list of possible labels for each of a list of aspects. A typical use case is aspect based sentiment analysis where one want to detect each aspect mentionned in a review along his assocated sentiment polarity (e.g., #product_quality: neutral, #delivery_time: negative, #price: positive, ...).\n* **Zero-shot classification** [Not yet implemented]: Assign one or more label to each text from a list of possible labels without the requirement of an annotated training dataset.\n\n`NLPipes` expose a simple `Model` API that offers a common abstraction to run several text classification tasks. The `Model` encapsulate most of the complex code from the library and save having to deal with the complexity of transformers based algorithms.\n\n#### Built with\n`NLPipes` is built with TensorFlow and HuggingFace Transformers:\n* [TensorFlow](https://www.tensorflow.org/): An end-to-end open source deep learning framework\n* [Transformers](https://huggingface.co/transformers/): An general-purpose open-sources library for transformers-based architectures\n\n## Getting Started\n\n#### Installation\n1. Create a virtual environment\n\n ```console\n python3 -m venv nlpipesenv\n source nlpipesenv/bin/activate\n ```\n\n2. Install the package\n\n ```console\n pip install nlpipes\n ```\n\n#### Examples\n\nHere are some use-cases with `NLPipes` on real datasets:\n\nName|Notebook|Description|Task|Size|Memory| \n----|-----------|-----|---------|---------|---------|\nGooglePlay Sentiment Detection|[Link]("")|Train a model to label reviews from the GooglePlay store reviews|Single-label classification|  |  \nStackOverflow tags Detection|Coming soon|Train a model to detect tags from the StackOverFlow questions |Multiple-labels classification|  | \nGooglePlay Aspect and Sentiment Detection|Coming soon|Train a model for aspect based sentiment detection to detect the aspects mentionned in GooglePlay reviews along his assocated sentiment polarity |Aspect-based classification|  | \n\n\n## Notice\n`NLPipes` is still in its early stage and is not yet suitable for production usage. Proceed with caution and use it at your own risk. The library comes with no warranty and future releases could bring substantial API and behavior changes.\n',
    'author': 'Ayhan UYANIK',
    'author_email': 'ayhan.uyanik@renault.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import transformers


class ModelCardConfig():
    """ This is the configuration class to store the configuration 
    of a `Model Card`. It is used to instantiate model card according
    to the specified arguments, describing the model's development
    and performances. """
    



import os
import logging

from typing import List
from typing import Optional
from typing import Union

from pathlib import Path

from dataclasses import dataclass

import tensorflow as tf
import transformers

from transformers import (
    PretrainedConfig,
    PreTrainedTokenizer,
    TFPreTrainedModel,
)

from nlpipes.trainers.trainers import Trainer

from nlpipes.losses.losses import softmax_cross_entropy

from nlpipes.data.data_loaders import DataLoader
from nlpipes.data.data_processors import DataProcessorForLanguageModeling

from nlpipes.pipelines.pipeline_utils import (
     select_corpus,
     augment_vocabulary,
)

from nlpipes.data.data_utils import (
     create_examples,
     split_examples,
)

from nlpipes.data.data_types import (
     Corpus,
     Token,
     InputExample,
)

from nlpipes.callbacks.callbacks import (
     Callback,
     TrainingStep,
     History,
     ModelCheckpoint,
     CSVLogger,
     ProgbarLogger,
     EarlyStopping,
     TimedStopping,
)

transformers.logging.set_verbosity_error()

logger = logging.getLogger('__name__')

#uncomment row below for debugging rust tokenizer
#os.environ["RUST_BACKTRACE"]="full" 


@dataclass
class TFPipelineForMaskedLM():
    """ The pipeline for masked language modeling task """
    
    model: TFPreTrainedModel
    tokenizer: PreTrainedTokenizer
    config: PretrainedConfig
    
    def __post_init__(self):
        self.sub_corpus: Corpus
    
    def train(
        self,
        texts: Corpus,
        train_frac: float=0.75,
        test_frac: float=0.25,
        num_epochs: int=2,
        batch_size: int=4,
        lr: float=2e-5,
        min_delta: float = 0.01,
        patience: int = 5,
        time_limit: int = 86400,
        rand_seed: int = 69,
        target_corpus_size: Union[int, float]=0.5,
        target_vocab_size: int=31000,
        similarity_metrics: Optional[List[str]]=['euclidean'],
        diversity_metrics: Optional[List[str]]=['entropy'],
        keep_checkpoints: bool = False,
        save_corpus_dir: str = './selected_corpus',
        checkpoints_dir: str = './checkpoints',
        logs_dir: str = './logs',
    ) -> TFPreTrainedModel:
        """ Train the model on a masked modeling task. """
        
        sub_corpus = select_corpus(
            corpus=texts,
            tokenizer=self.tokenizer,
            target_corpus_size=target_corpus_size,
            similarity_metrics=similarity_metrics,
            diversity_metrics=diversity_metrics,
        )
        
        new_tokens = augment_vocabulary(
            corpus=sub_corpus,
            tokenizer=self.tokenizer,
            target_vocab_size=target_vocab_size,
        )
        
        self.sub_corpus = sub_corpus
        self.tokenizer.add_tokens(new_tokens)
        self.model.resize_token_embeddings(len(self.tokenizer))
        
        examples = create_examples(
            texts=sub_corpus, 
            labels=None,
            config=self.config,
        )
        
        train_examples, test_examples = split_examples(
            examples=examples,
            train_frac=train_frac,
            test_frac=test_frac, 
            shufffle=True, 
        )

        data_processor = DataProcessorForLanguageModeling(
            name='DataProcessor',
            tokenizer=self.tokenizer,
            config=self.config,
            mlm_proba=0.15
        )
        
        train_stream = DataLoader(
            name='TrainDataLoader',
            examples=train_examples,
            batch_size=batch_size,
            data_processor=data_processor,
        )
        
        test_stream = DataLoader(
            name='TestDataLoader',
            examples=test_examples,
            batch_size=batch_size,
            data_processor=data_processor,
        )
        
        training_step = TrainingStep(
            name='BatchTraining',
            model=self.model,
            loss_function=softmax_cross_entropy,
            optimizer=tf.optimizers.Adam(lr),
        )
        
        history = History(
            name='History', 
            training_step=training_step,
            metric=tf.metrics.Mean,
        )
        
        model_checkpoint = ModelCheckpoint(
            name='ModelCheckpoint',
            model=self.model, 
            history=history,
            checkpoints_dir=checkpoints_dir,
        )
        
        early_stopping = EarlyStopping(
            name='EarlyStopping', 
            history=history,
            direction='minimize',
            min_delta=min_delta,
            patience=patience,
        )
        
        timed_stopping = TimedStopping(
            name='TimedStopping',
            time_limit=time_limit,
        )
       
        csv_logger = CSVLogger(
            name='CSVLogging',
            logs_dir=logs_dir,
        )
        
        progbar_logger = ProgbarLogger(
            name='ProgbarLogger',
            batch_size=batch_size,
            num_epochs=num_epochs,
            history=history,
            num_samples=sum([
                train_stream.num_examples(),
                test_stream.num_examples(),
            ])
        )
        
        callbacks = [
            training_step,
            history,
            model_checkpoint,
            csv_logger,
            progbar_logger,
            early_stopping,
            timed_stopping,
        ]
        
        trainer = Trainer(
            train_stream=train_stream,
            test_stream=test_stream,
            callbacks=callbacks,
            num_epochs=num_epochs,
        )
        
        trainer.train()
        
        if keep_checkpoints == False:
            shutil.rmtree(checkpoints_dir)
        
        return history
    
    def evaluate(self):
        raise NotImplementedError
        
    def predict(self):
        raise NotImplementedError
    
    def save(self, save_dir: str):
        """ Save the model and the tokenizer on disk """
        
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)
            
        corpus_path = os.path.join(save_dir, 'selected_corpus.txt')         
        Path(corpus_path).write_text('\n'.join(self.sub_corpus))
        self.tokenizer.save_pretrained(save_dir)
        self.model.save_pretrained(save_dir)
        